from fipy import CellVariable, TransientTerm, DiffusionTerm
from fipy.meshes.cylindricalNonUniformGrid1D import CylindricalNonUniformGrid1D
import numpy as np
from h5py import File
from os.path import exists
import sys

solve_coupled = True
solve_uncoupled = True

if solve_coupled:
    print ('SOLVING COUPLED PROBLEM')
else:
    print ('_NOT_ SOLVING COUPLED PROBLEM')
    
if solve_uncoupled:
    print ('SOLVING EIGENVALUE UNCOUPLED PROBLEM')
else:
    print ('_NOT_ SOLVING EIGENVALUE UNCOUPLED PROBLEM')

def save_hdf5(h,sec):
    # paraview doesn't like leading integer in label
    ts = 'Time:   %.8e s/' % (sec)
    
    # require_dataset creates dataset if it doesn't exist (create_dataset throws error), requiring type & shape
    if solve_coupled:
        h.require_dataset(ts + 'coupled_pD',          data=pD.value[:,None,None],          dtype='float64', shape=(nx,1,1))
        h.require_dataset(ts + 'coupled_pD_grad_X',   data=pD.grad.value[0,:,None,None],   dtype='float64', shape=(nx,1,1))
        h.require_dataset(ts + 'coupled_psiD',        data=psiD.value[:,None,None],        dtype='float64', shape=(nx,1,1))
        h.require_dataset(ts + 'coupled_psiD_grad_X', data=psiD.grad.value[0,:,None,None], dtype='float64', shape=(nx,1,1))

    if solve_uncoupled:
        h.require_dataset(ts + 'delta1',                data=delta1.value[:,None,None],        dtype='float64', shape=(nx,1,1))
        h.require_dataset(ts + 'delta1_grad_X',         data=delta1.grad.value[0,:,None,None], dtype='float64', shape=(nx,1,1))
        h.require_dataset(ts + 'delta2',                data=delta2.value[:,None,None],        dtype='float64', shape=(nx,1,1))
        h.require_dataset(ts + 'delta2_grad_X',         data=delta2.grad.value[0,:,None,None], dtype='float64', shape=(nx,1,1))
        h.require_dataset(ts + 'uncoupled_pD',          data=UpD.value[:,None,None],           dtype='float64', shape=(nx,1,1))
        h.require_dataset(ts + 'uncoupled_pD_grad_X',   data=UpD.grad.value[0,:,None,None],    dtype='float64', shape=(nx,1,1))
        h.require_dataset(ts + 'uncoupled_psiD',        data=UpsiD.value[:,None,None],         dtype='float64', shape=(nx,1,1))
        h.require_dataset(ts + 'uncoupled_psiD_grad_X', data=UpsiD.grad.value[0,:,None,None],  dtype='float64', shape=(nx,1,1))

# compare solution of coupled and eigenvalue-uncoupled 1D problems

# time in seconds
SECINHR = 60.0*60.0
SECINDAY = SECINHR*24.0
SECINYR = SECINDAY*365.25  # NB: pflotran assumes 365.0

nx = 550   # plot_fipy_results.py assumes more than 175 spatial nodes 
dx0 = 0.001
r = 1.02
nt = 500   # plot_fipy_results.py assumes more than 300 times

dtmin = 1.0E-14
dtmax = 1.5

alphaD_vec = [1.0E+2, 1.0E+5, 1.0E+8]
KD_vec = [1.0E-1, 1.0E-4, 1.0E-7]

idx = np.arange(nx)
dx = dx0*r**idx  # create large domain with fine mesh at origin
L = dx.sum()
origin = dx0

print ('domain size L=',L,' nx=',nx,' dx0=',dx0,' r=',r)
print ('dx:',dx[:5],'...',dx[-5:])

#alphaD = alphaD_vec[1]
#KD = KD_vec[1]

# both positive, both negative, +psi & -p 

gradpsi = -float(sys.argv[1]) 
gradp   = -float(sys.argv[2]) 

for ii,alphaD in enumerate(alphaD_vec):
    for jj,KD in enumerate(KD_vec):

        print (ii,'alphaD=',alphaD,'::',jj,'KD=',KD)
        
        mesh = CylindricalNonUniformGrid1D(dx=dx, nx=nx, origin=(origin,))

        if solve_coupled:
            ## ========================================================================
            ## solve coupled problem directly

            # dimensionless change in electrostatic potential
            psiD = CellVariable(name='$\\psi_D$', mesh=mesh, value=0.0)

            psiD.faceGrad.constrain(gradpsi, mesh.facesLeft) # type II at wellbore
            psiD.constrain(0.0, mesh.facesRight) # homogeneous type I in far-field
            
            # dimensionless pressure change
            pD = CellVariable(name='$p_D$', mesh=mesh, value=0.0)

            pD.faceGrad.constrain(gradp, mesh.facesLeft) # type II at wellbore
            pD.constrain(0.0, mesh.facesRight) # homogeneous type I in far-field
                
            # coupled dimensionless equations
            c_electrical = (TransientTerm(1.0, var=psiD) ==
                            DiffusionTerm(alphaD, var=psiD) + DiffusionTerm(alphaD, var=pD))

            c_hydraulic = (TransientTerm(1.0, var=pD) ==
                           DiffusionTerm(KD, var=psiD) + DiffusionTerm(1.0, var=pD))

            coupled_eq = c_electrical & c_hydraulic # coupled
            
            print ('coefficients in coupled equations:')
            print ('[%.8E , %.8E]' % (alphaD,alphaD))
            print ('[%.8E , %.8E]' % (KD,1.0))

            print ('boundary conditions in coupled problem:')
            print ('grad psi = %.8E' % (gradpsi,))
            print ('grad p   = %.8E' % (gradp,))

        if solve_uncoupled:
            # $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
            # solve eigenvalue uncoupled problem for comparison 

            delta1 = CellVariable(name='$\\delta_1$', mesh=mesh, value=0.0)
            delta2 = CellVariable(name='$\\delta_2$', mesh=mesh, value=0.0)

            UpsiD = CellVariable(name='uncoupled_psiD', mesh=mesh, value=0.0)
            UpD   = CellVariable(name='uncoupled_pD',   mesh=mesh, value=0.0)
            
            # Delta
            Delta = np.sqrt(1.0 + alphaD * (4.0*KD - 2.0 + alphaD))
            Gpp = (1.0 - (alphaD + Delta))/2.0
            Gp =  -KD*alphaD / Gpp

            Lambda2 = (1.0 + alphaD + Delta)/2.0
            Lambda1 = alphaD*(1.0 - KD)/Lambda2

            print ('Delta:',Delta,'Gp:',Gp,'Gpp:',Gpp)
            print ('coefficients in uncoupled equations')
            print ('[%.8E , %.8E]' % (Lambda1,0.0))
            print ('[%.8E , %.8E]' % (0.0,Lambda2))
            
            Q1 = -(Gpp * gradp + KD * gradpsi)/Delta
            delta1.faceGrad.constrain(Q1, mesh.facesLeft) # type II at wellbore
            delta1.constrain(0.0, mesh.facesRight) # homogeneous type I in far-field

            Q2 =  (Gp * gradp + KD * gradpsi)/Delta
            delta2.faceGrad.constrain(Q2, mesh.facesLeft) # type II at wellbore
            delta2.constrain(0.0, mesh.facesRight) # homogeneous type I in far-field

            print ('boundary conditions in uncoupled problem:')
            print ('Q1 = %.8E' % (Q1,))
            print ('Q2 = %.8E' % (Q2,))

            # specificataion of un-coupled dimensionless equations
            U_eq1 = (TransientTerm(1.0, var=delta1) == DiffusionTerm(Lambda1, var=delta1))
            U_eq2 = (TransientTerm(1.0, var=delta2) == DiffusionTerm(Lambda2, var=delta2))

        # save initial state into hdf5 file
        # ====================================================

        fn = 'theis-fipy-output-%i-%i.h5' % (ii,jj)

        if exists(fn):
            EXISTS = True
            fh5 = File(fn ,'r+')
        else:
            EXISTS = False
            fh5 = File(fn ,'w')            

        if not EXISTS:
            # create face-centered coordinates, as expected by paraview
            Xc = mesh.cellCenters.value[0,:]
            Xf = np.empty((nx+1),dtype=np.float64)
            Xf[1:-1] = (Xc[:-1] + Xc[1:])/2.0
            Xf[0] = Xc[0] - (Xf[1] - Xc[0])
            Xf[-1] = Xc[-1] + (Xc[-1] - Xf[-2])
                
            # other dimensions in 1D problem are arbitrary,
            # but make them smaller than length to make it look "longer" in x direction
            # (this is just for asthetics)
            unit = np.array([0.0,L/10.0],dtype=np.float64) 

            fh5.create_dataset('Coordinates/X [m]',data=Xf)  # face centered (for paraview)
            fh5.create_dataset('Coordinates/Xc [m]',data=Xc) # cell centered (for my plotting script)
            fh5.create_dataset('Coordinates/Y [m]',data=unit)
            fh5.create_dataset('Coordinates/Z [m]',data=unit)

            prov = fh5.create_group('Provenance')
            param = fh5.create_group('Parameters')
            param.attrs['KD'] = KD 
            param.attrs['alphaD'] = alphaD
            param.attrs['pressure_BC_[Pa|m]'] = gradp
            param.attrs['electrostatic_BC_[V|m]'] = gradpsi

        if solve_uncoupled:
            # compute uncoupled physical solution
            # from intermediate variables
            UpsiD.setValue(-(delta1.value*Gp + delta2.value*Gpp)/KD)
            UpD.setValue(delta1.value + delta2.value)

        t = 0.0
        save_hdf5(fh5,t)

        # solve both systems
        # ==============================
        
        for j,step in enumerate(np.concatenate((np.logspace(np.log10(dtmin),np.log10(dtmax),nt/2),
                                                np.ones((nt/2,))*dtmax),axis=0)):
        #for j,step in enumerate(np.ones(nt,)*1.5E-1):
            t += step

            if solve_coupled:
                # solve coupled equations directly
                coupled_eq.solve(dt=step)
                    
            if solve_uncoupled:
                # solve uncoupled equations
                U_eq1.solve(dt=step)
                U_eq2.solve(dt=step)
                        
                # compute uncoupled physical solution
                # from intermediate variables
                UpsiD.setValue(-(delta1.value*Gp + delta2.value*Gpp)/KD)
                UpD.setValue(delta1.value + delta2.value)
            
            save_hdf5(fh5,t)
            if j%50 == 0:
                print j,'t=',t
            
        fh5.close()
        print ('<done>',ii,jj)
