import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from h5py import File

figsize = (8.5,6.5)
#ymin = 1.0E-24

# these axes are for plots across parameters/model realizations
PARFIG = plt.figure(5,figsize=figsize) # across-parameters.png
PARAXES = []
for j in range(6):
    PARAXES.append(PARFIG.add_subplot(2,3,j+1))

DIFFIG = plt.figure(4,figsize=(7,11)) # coupled/uncoupled difference
DIFAXES = [DIFFIG.add_subplot(2,1,1), DIFFIG.add_subplot(2,1,2)]
    
# what to plot (6 subplots)
plt_vars = ['coupled_pD',  'delta1','uncoupled_pD',
            'coupled_psiD','delta2','uncoupled_psiD']

plt_var_names = ['coupled $p_D$',  '$\\delta_1$','eig. uncoupled $p_D$',
                 'coupled $\\psi_D$','$\\delta_2$','eig. uncoupled $\\psi_D$']

lts = ['-','--',':','-.'] # alphaD (shades are KD)
black_colors = ['black','dimgray','darkgray','lightgray'] # p
blue_colors = ['midnightblue','blue','deepskyblue'] # delta1/2
red_colors = ['red','salmon','lightsalmon'] # psi
lw=1.5
xloc = 150  # << change 150 -> 100

colors = [black_colors, blue_colors, black_colors,
          red_colors, blue_colors, red_colors]

# alphaD loop
for ii in [0,1,2]:
    # KD loop
    for jj in [0,1,2]:

        fh = File('theis-fipy-output-%i-%i.h5' % (ii,jj),'r')

        alphaD = fh['Parameters'].attrs['alphaD']
        KD = fh['Parameters'].attrs['KD']
        var_string = "  $\\alpha_D=10^{%i}$ $K_D=10^{%i}$" % (np.log10(alphaD),np.log10(KD))
        
        print (ii,'alphaD=',alphaD,'::',jj,'KD=',KD)
        
        Xc = fh['Coordinates']['Xc [m]'][:] # cell centered (computed in other script)
        print ('plotting location:',xloc,Xc[xloc])
        
        time_keys = [x for x in fh.keys() if 'Time:' in x]
        times = np.array([float(x.split(':')[1].split()[0]) for x in time_keys])

        nt = times.shape[0] 
        sort_idx = np.argsort(times)
        initial = time_keys[sort_idx[0]] # t=0
        var_names = sorted(fh[initial].keys())
        nvars = len(var_names)
        var = list(fh[initial].keys())[0] # first variable
        nx = fh[initial][var][:,0,0].shape[0] # domain size

        d = np.empty((nx,nt,nvars),dtype=np.float32)

        for i in range(nt):
            for j in range(nvars):
                d[:,i,j] = fh[time_keys[sort_idx[i]]][var_names[j]][:,0,0]        
        
        vars_idx = []
        for pv in plt_vars:
            for j,vn in enumerate(var_names):
                if pv == vn:
                    vars_idx.append(j)
                    break

        # plot variation in time at some spaces (for a single set of parameters)
        for i,PAX in enumerate(PARAXES):
            tt = times[sort_idx]
            # /Xc[xloc]**2
            PARAXES[i].loglog(tt,d[xloc,:,vars_idx[i]]/np.sqrt(Xc[xloc]),
                              lw=lw, ls=lts[ii], color=colors[i][jj])
            
            PAX.set_title(var_names[vars_idx[i]] + var_string)
            #PAX.set_ylim(ymin=ymin)

        fh.close()

        # compare pD solutions (coupled/uncoupled)
        pC = vars_idx[0]
        pU = vars_idx[2]
        diff = d[xloc,:,pC] - d[xloc,:,pU]
        rel = diff/d[xloc,:,pC]        
        DIFAXES[0].loglog(tt/Xc[xloc]**2, np.abs(diff), '-',
                          label='$10^{%i}$ $10^{%i}$' % (np.log10(alphaD),np.log10(KD)))
        print('  p difference= rel min:',np.nanmin(rel),' rel max:',np.nanmax(rel),' # non-zero:',np.sum(diff>0))

        # compare psiD solutions
        psiC = vars_idx[3]
        psiU = vars_idx[5]
        diff = d[xloc,:,psiC] - d[xloc,:,psiU]
        rel = diff/d[xloc,:,psiC]
        DIFAXES[1].loglog(tt/Xc[xloc]**2, np.abs(diff), '-',
                          label='$10^{%i}$ $10^{%i}$' % (np.log10(alphaD),np.log10(KD)))
        print('  psi difference= rel min:',np.nanmin(rel),' rel max:',np.nanmax(rel),' # non-zero:',np.sum(diff>0))

DIFAXES[0].set_xlabel('$t/r^2$')
DIFAXES[0].set_ylabel('$p_D$: |Coupled - Uncoupled|')
DIFAXES[0].legend(loc=0,ncol=2,fontsize='x-small')
DIFAXES[1].set_xlabel('$t/r^2$')
DIFAXES[1].set_ylabel('$\\psi_D$: |Coupled - Uncoupled|')
DIFAXES[1].legend(loc=0,ncol=2,fontsize='x-small')

for i,letter in enumerate(["a","b","c","d","e","f"]):
    PARAXES[i].set_yscale('symlog',linthreshy=1.0E-22)
    PARAXES[i].set_ylim([-1,1])
    PARAXES[i].set_xlim([1.0E-13,100])
    PARAXES[i].set_title(plt_var_names[i])
    PARAXES[i].set_yticks([-1.0,-1.0E-4,-1.0E-8,-1.0E-12,-1.0E-16,0.0,1.0E-16,1.0E-12,1.0E-8,1.0E-4,1.0])
    PARAXES[i].text(1.0E-12,1.0E-4,f"{letter})",size=14)

for i in range(6):
    PARAXES[i].axhline(0.0,color='gray',lw=0.25,zorder=-100)
    if i in [0,3]:
        # left-most plots        
        PARAXES[i].set_ylabel('$y/r^{1/2}$')
    else:
        PARAXES[i].set_ylabel('')
        PARAXES[i].set_yticklabels([])
    
for i in range(6):
    if i in [3,4,5]:
        # bottom row plots
        PARAXES[i].set_xlabel('$t/r^2$')
    else:
        PARAXES[i].set_xlabel('')
        PARAXES[i].set_xticklabels([])

#plot([0.5, 0.5], [0, 1], color='lightgreen', lw=5,transform=gcf().transFigure, clip_on=False)
PARAXES[5].plot([0.395, 0.395], [0.02, 0.98], color='gray', lw=2.0, 
         transform=PARFIG.transFigure, clip_on=False)

        
PARFIG.tight_layout()
PARFIG.savefig('across-parameters.png',dpi=200)
plt.close(5)

DIFFIG.tight_layout()
DIFFIG.savefig('coupled-uncoupled-comparison.png')
plt.close(4)

